from __future__ import annotations

from abc import abstractmethod
from io import BytesIO
from math import exp, log
from pathlib import Path
from typing import TYPE_CHECKING, Sequence, cast

import matplotlib.pyplot as plt  # type: ignore
from matplotlib.backend_bases import MouseEvent as PlotMouseEvent  # type: ignore
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg  # type: ignore
from matplotlib.figure import Figure  # type: ignore
from matplotlib.layout_engine import ConstrainedLayoutEngine  # type: ignore
from matplotlib.lines import Line2D  # type: ignore
from matplotlib.rcsetup import cycler  # type: ignore
from PyQt6.QtCore import QEvent, QPointF, Qt, pyqtSignal
from PyQt6.QtGui import QImage, QMouseEvent, QNativeGestureEvent, QWheelEvent
from PyQt6.QtWidgets import QFileDialog, QFrame

if TYPE_CHECKING:
    from ...main import MainWindow
    from ..types import Frame

STYLE_DIR = Path(__file__).parent / 'plotting.mplstyle'


__all__ = [
    'apply_plotting_style',

    'PlottingCanvas',

    'PlotMouseEvent'
]


def apply_plotting_style() -> None:
    plt.style.use({
        'axes.edgecolor': '#FFFFFF3D',
        'axes.facecolor': '#FFFFFF07',
        'axes.labelcolor': '#FFFFFFD9',
        'axes.prop_cycle': cycler('color', [
            '#FF6200', '#696969', '#525199', '#60A6DA', '#D0D93C', '#A8A8A9', '#FF0000', '#349651', '#AB0066'
        ]),
        'figure.facecolor': '#19232D',
        'legend.edgecolor': '#FFFFFFD9',
        'legend.facecolor': 'inherit',
        'legend.framealpha': 0.12,
        'markers.fillstyle': 'full',
        'savefig.facecolor': '#19232D',
        'text.color': 'white',
        'xtick.color': '#FFFFFFD9',
        'ytick.color': '#FFFFFFD9'
    })


class PlottingCanvas(FigureCanvasQTAgg):
    WHEEL_STEP = 15 * 8  # degrees

    mouseMoved = pyqtSignal(QMouseEvent)
    wheelScrolled = pyqtSignal(int)

    def __init__(
        self, main: MainWindow, ylog: bool = False, xlog: bool = False, controls: bool = True,
        xpad: float | tuple[float, float] | None = None, ypad: float | tuple[float, float] | None = None,
        figsize: tuple[int, int] = (5, 4)
    ) -> None:
        from ..abstracts import HBoxLayout, PushButton
        from ..types import Stretch

        self.main = main

        self.ylog, self.xlog = ylog, xlog

        self.figure = Figure(figsize, self.main.settings.base_ppi)

        self.axes = self.figure.add_subplot(111)

        self.figure.set_layout_engine(ConstrainedLayoutEngine())

        super().__init__(self.figure)

        self.mpl_connect('motion_notify_event', self._on_mouse_moved)

        self.angleRemainder = 0
        self.zoomValue = 0.0

        self.clicked = False
        self.old_pos = QPointF(0.0, 0.0)

        self.controls = QFrame()

        self.copy_frame_button = PushButton('Copy Graph', clicked=self.copy_graph_to_clipboard)

        self.save_frame_as_button = PushButton('Save Graph as', clicked=self.on_save_graph_as_clicked)

        _controls = [self.copy_frame_button, self.save_frame_as_button, Stretch()]

        if controls:
            self.controls = HBoxLayout(_controls)
        else:
            self.controls = QFrame()
            HBoxLayout(self.controls, _controls)
            self.controls.hide()

        self.xpad = (xpad, xpad) if (not isinstance(xpad, tuple) and xpad is not None) else xpad
        self.ypad = (ypad, ypad) if (not isinstance(ypad, tuple) and ypad is not None) else ypad

    def _on_mouse_moved(self, event: PlotMouseEvent) -> None:
        if not event.inaxes or not self.axes.lines:
            return

        if event.xdata is None or event.ydata is None:
            return

        if event.xdata < self.line._xorig[0] or event.xdata > self.line._xorig[-1]:
            return

        self.on_mouse_moved(event)

    def on_mouse_moved(self, event: PlotMouseEvent) -> None:
        ...

    @classmethod
    def limits_to_range(cls, lim: tuple[int, int]) -> int:
        return lim[1] - lim[0]

    @property
    def line(self) -> Line2D:
        return cast(Line2D, self.axes.lines[0])

    @property
    def cur_lims(self) -> tuple[tuple[float, float], tuple[float, float]]:

        try:
            return (
                tuple[float, float](log(x) if self.xlog else x for x in self.axes.get_xlim()),
                tuple[float, float](log(x) if self.ylog else x for x in self.axes.get_ylim())
            )
        except ValueError:
            return (self.axes.get_xlim(), self.axes.get_ylim())

    @cur_lims.setter
    def cur_lims(self, new_lims: tuple[Sequence[float], Sequence[float]]) -> None:
        new_xlim, new_ylim = new_lims

        try:
            self.axes.set_xlim(tuple[float, float](exp(x) if self.xlog else x for x in new_xlim))
            self.axes.set_ylim(tuple[float, float](exp(x) if self.ylog else x for x in new_ylim))

            self.figure.canvas.draw_idle()
        except OverflowError:
            # zooming/panning too much
            ...

    def event(self, event: QEvent) -> bool:
        if isinstance(event, QNativeGestureEvent):
            typ = event.gestureType()

            if typ == Qt.NativeGestureType.BeginNativeGesture:
                self.zoomValue = 0.0
            elif typ == Qt.NativeGestureType.ZoomNativeGesture:
                self.zoomValue += event.value()

            if typ == Qt.NativeGestureType.EndNativeGesture:
                self.zoom_function(event, -1 if self.zoomValue < 0 else 1)

        return super().event(event)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if event.button() == Qt.MouseButton.LeftButton:
            self.clicked = True

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        self.clicked = False

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        self.mouseMoved.emit(event)

        if self.clicked:
            cur_xlim, cur_ylim = self.cur_lims

            curr_pos = event.globalPosition()

            rel_x = self.old_pos.x() - curr_pos.x()
            rel_y = self.old_pos.y() - curr_pos.y()

            width, height = self.get_width_height()
            v_width, v_height = cur_xlim[1] - cur_xlim[0], cur_ylim[1] - cur_ylim[0]

            xrate = rel_x * v_width / width
            yrate = rel_y * v_height / height

            self.cur_lims = (
                (x + xrate for x in cur_xlim),
                (x - yrate for x in cur_ylim)
            )

        self.old_pos = event.globalPosition()

        try:
            super().mouseMoveEvent(event)
        except Exception:
            event.ignore()

    def wheelEvent(self, event: QWheelEvent) -> None:
        modifier = event.modifiers()
        mouse = event.buttons()

        if modifier == Qt.KeyboardModifier.ControlModifier or mouse in {
            Qt.MouseButton.RightButton, Qt.MouseButton.MiddleButton
        }:
            angleDelta = event.angleDelta().y()

            if self.angleRemainder * angleDelta < 0:
                self.angleRemainder = 0

            self.angleRemainder += angleDelta

            if abs(self.angleRemainder) >= self.WHEEL_STEP:
                steps = self.angleRemainder // self.WHEEL_STEP
                self.zoomValue += steps
                self.zoom_function(event, steps)

                self.angleRemainder %= self.WHEEL_STEP

        event.ignore()

    def zoom_function(self, event: QWheelEvent, steps: int) -> None:
        if not steps:
            return

        scale_factor = 1
        base_scale = 1.1

        for _ in range(abs(steps)):
            if steps > 0:
                scale_factor *= base_scale
            else:
                scale_factor /= base_scale

        cur_xlim, cur_ylim = self.cur_lims

        pos = event.position()
        xdata, ydata = pos.x(), pos.y()
        width, height = self.get_width_height()

        xrate, yrate = xdata / width, ydata / height

        xabs = cur_xlim[1] * xrate + cur_xlim[0] * (1 - xrate)
        yabs = cur_ylim[0] * yrate + cur_ylim[1] * (1 - yrate)

        new_width = abs(cur_xlim[1] - cur_xlim[0]) / scale_factor
        new_height = abs(cur_ylim[1] - cur_ylim[0]) / scale_factor

        self.cur_lims = (
            (xabs - new_width * xrate, xabs + new_width * (1 - xrate)),
            (yabs - new_height * (1 - yrate), yabs + new_height * yrate)
        )

    def clear(self) -> None:
        self.axes.clear()

        if self.ylog:
            self.axes.set_yscale("log")

        if self.xlog:
            self.axes.set_xscale("log")

    def draw(self) -> None:
        if 0 not in self.figure.get_size_inches():
            super().draw()

    def copy_graph_to_clipboard(self) -> None:
        buffer = BytesIO()

        self.figure.savefig(buffer, transparent=False)

        self.main.clipboard.setImage(QImage.fromData(buffer.getvalue()))

        buffer.close()

        self.main.show_message('Graph successfully copied to clipboard')

    def on_save_graph_as_clicked(self, checked: bool | None = None) -> None:
        save_path_str, _ = QFileDialog.getSaveFileName(
            self.main, 'Save as', f'graph_{self.main.current_output.last_showed_frame}', 'Graph (*.svg)'
        )

        self.figure.savefig(save_path_str, transparent=False)

    @abstractmethod
    def _render(self, frame: Frame) -> None:
        raise NotImplementedError

    def render(self, frame: Frame, set_lims: bool = True) -> None:
        if not set_lims:
            self.zoomValue = 0.0

        lims = self.cur_lims if self.zoomValue else None

        self.clear()

        self._render(frame)

        self.axes.legend()
        self.figure.canvas.draw_idle()

        if lims:
            self.cur_lims = lims
        elif self.xpad or self.ypad:
            xlim, ylim = self.cur_lims

            line = self.line

            self.cur_lims = (
                (line._xorig[0] - self.xpad[0], line._xorig[-1] + self.xpad[1]) if self.xpad else xlim,
                (line._yorig[0] - self.ypad[0], line._yorig[-1] + self.ypad[1]) if self.ypad else ylim
            )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from PyQt6.QtCore import QKeyCombination, Qt
from PyQt6.QtGui import QCloseEvent
from PyQt6.QtWidgets import QLabel, QTabWidget, QWidget

from ..core import ExtendedDialog, HBoxLayout, PushButton, VBoxLayout

if TYPE_CHECKING:
    from .window import MainWindow


__all__ = [
    'ScriptErrorDialog',
    'SettingsDialog'
]


class ScriptErrorDialog(ExtendedDialog):
    __slots__ = ('main', 'label', 'reload_button', 'exit_button')

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__(main_window, Qt.WindowType.Dialog)
        self.main = main_window

        self.setWindowTitle('There was an error')
        self.setModal(True)

        self.setup_ui()
        self.setup_shortcuts()

        self.set_qobject_names()

    def setup_ui(self) -> None:
        self.vlayout = VBoxLayout(self)

        self.label = QLabel()

        self.reload_button = PushButton('Reload', self, clicked=self.on_reload_clicked)
        self.exit_button = PushButton('Exit', self, clicked=self.on_exit_clicked)

        self.vlayout.addWidget(self.label)
        self.vlayout.addLayout(HBoxLayout([
            self.reload_button, self.exit_button
        ]))

    def setup_shortcuts(self) -> None:
        self.add_shortcut(QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_R).toCombined(), self.reload_button.click)

    def on_reload_clicked(self, clicked: bool | None = None) -> None:
        self.hide()
        self.main.reload_script()

    def on_exit_clicked(self, clicked: bool | None = None) -> None:
        self.hide()
        self.script_exec_failed = True

        if not self.main.no_exit:
            sys.exit(1)

    def closeEvent(self, event: QCloseEvent) -> None:
        self.on_exit_clicked()


class SettingsDialog(ExtendedDialog):
    __slots__ = ('main', 'tab_widget',)

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__(main_window)

        self.main = main_window
        self.setWindowTitle('Settings')

        self.setup_ui()

        self.set_qobject_names()

    def setup_ui(self) -> None:
        self.tab_widget = QTabWidget(self)
        self.vlayout = VBoxLayout(self, [self.tab_widget])

    def addTab(self, widget: QWidget, label: str) -> int:
        return self.tab_widget.addTab(widget, label)



"""Previewer for VapourSynth scripts"""

__version__ = '0.7.0'

__author__ = 'Endilll <>'
__maintainer__ = 'Setsugen no ao <setsugen@setsugen.dev>'

__author_name__, __author_email__ = [x[:-1] for x in __author__.split('<')]
__maintainer_name__, __maintainer_email__ = [x[:-1] for x in __maintainer__.split('<')]

if __name__ == '__github__':
    print(__version__)



from __future__ import annotations

import logging
import sys
from typing import Callable

from vapoursynth import MessageType

__all__ = [
    'setup_logger', 'set_log_level',

    'get_vs_logger',
]

_logger_setup = False


LOG_LEVEL = logging.INFO


_vsLogType_to_logging = {
    MessageType.MESSAGE_TYPE_DEBUG: logging.DEBUG,
    MessageType.MESSAGE_TYPE_INFORMATION: logging.INFO,
    MessageType.MESSAGE_TYPE_WARNING: logging.WARNING,
    MessageType.MESSAGE_TYPE_CRITICAL: logging.CRITICAL,
    MessageType.MESSAGE_TYPE_FATAL: logging.FATAL,
}


def setup_logger() -> None:
    global _logger_setup

    if not _logger_setup:
        logging.basicConfig(format='{asctime}: {name}: {levelname}: {message}', style='{', level=LOG_LEVEL)
        logging.Formatter.default_msec_format = '%s.%03d'
        if sys.stdout.isatty():
            logging.addLevelName(
                logging.DEBUG, "\033[0;32m%s\033[0m" % logging.getLevelName(logging.DEBUG)
            )
            logging.addLevelName(
                logging.INFO, "\033[1;33m%s\033[0m" % logging.getLevelName(logging.INFO)
            )
            logging.addLevelName(
                logging.WARNING, "\033[1;35m%s\033[1;0m" % logging.getLevelName(logging.WARNING)
            )
            logging.addLevelName(
                logging.ERROR, "\033[1;41m%s\033[1;0m" % logging.getLevelName(logging.ERROR)
            )

        _logger_setup = True


def set_log_level(main: int = LOG_LEVEL, engine: int = logging.ERROR) -> None:
    from vsengine import _hospice  # type: ignore[import]
    from vsengine import policy

    policy.logger.addFilter(lambda record: 'dead environment' not in record.msg)
    _hospice.logger.setLevel(engine)
    logging.getLogger().level = main


def get_vs_logger() -> Callable[[MessageType, str], None]:
    setup_logger()

    def _logger(mType: MessageType, msg: str) -> None:
        import logging
        return logging.log(_vsLogType_to_logging[mType], msg)

    return _logger



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Iterator, Mapping, cast, overload

from ..core import AbstractToolbar, AbstractToolbarSettings, AbstractYAMLObjectSingleton, storage_err_msg
from .main import MainToolbar
from .playback import PlaybackToolbar

if TYPE_CHECKING:
    from ..main import MainWindow
    from .benchmark import BenchmarkToolbar
    from .comp import CompToolbar
    from .debug import DebugToolbar
    from .misc import MiscToolbar
    from .pipette import PipetteToolbar
    from .scening import SceningToolbar


__all__ = [
    'MainToolbar',

    'Toolbars'
]


class Toolbars(AbstractYAMLObjectSingleton):
    __slots__ = ()

    main: MainToolbar
    playback: PlaybackToolbar
    scening: SceningToolbar
    pipette: PipetteToolbar
    benchmark: BenchmarkToolbar
    misc: MiscToolbar
    comp: CompToolbar
    debug: DebugToolbar

    _closure = {**globals()}

    @classmethod
    def name_to_tooltype(cls, name: str) -> type[AbstractToolbar]:
        exec(f'from .{name} import {name.title()}Toolbar as _inner_tb_{name}', cls._closure)
        return cls._closure[f'_inner_tb_{name}']  # type: ignore # noqa

    def __init__(self, main_window: MainWindow) -> None:
        main_window.toolbars = self

        self.main = MainToolbar(main_window)
        self.playback = PlaybackToolbar(main_window)

        self.toolbar_names = [
            'debug', 'scening', 'pipette', 'benchmark', 'misc', 'comp'
        ]

        self.toolbars = dict(main=self.main, playback=self.playback) | {
            name: self.name_to_tooltype(name)(main_window)
            for name in self.toolbar_names
        }

        for name, toolbar in self.toolbars.items():
            toolbar.setObjectName(f'Toolbars.{name}')

            if not hasattr(self, name):
                setattr(self, name, toolbar)

    @overload
    def __getitem__(self, _sub: str | int) -> AbstractToolbar:
        ...

    @overload
    def __getitem__(self, _sub: slice) -> list[AbstractToolbar]:
        ...

    def __getitem__(self, _sub: str | int | slice) -> AbstractToolbar | list[AbstractToolbar]:
        length = len(self.toolbars)
        if isinstance(_sub, slice):
            return [self[i] for i in range(*_sub.indices(length))]

        if isinstance(_sub, int):
            if _sub < 0:
                _sub += length
            if _sub < 0 or _sub >= length:
                raise IndexError

            _sub = list(self.toolbars.keys())[_sub]

        return cast(AbstractToolbar, getattr(self, _sub))

    def __len__(self) -> int:
        return len(self.toolbars)

    if TYPE_CHECKING:
        # https://github.com/python/mypy/issues/2220
        def __iter__(self) -> Iterator[AbstractToolbar]:
            ...

    def should_set_state(self, cls: type[AbstractToolbar] | type[AbstractToolbarSettings]) -> bool:
        if issubclass(cls, AbstractToolbarSettings):
            name = cls.__name__[:-8]
        else:
            name = cls.__name__[:-7]

        return name.lower() in self.toolbars

    def __getstate__(self) -> Mapping[str, Mapping[str, Any]]:
        return {
            toolbar_name: getattr(self, toolbar_name).__getstate__()
            for toolbar_name in self.toolbars
        }

    def __setstate__(self, state: Mapping[str, Mapping[str, Any]]) -> None:
        for toolbar_name in self.toolbars:
            try:
                storage = state[toolbar_name]
                if not isinstance(storage, Mapping):
                    raise TypeError
                getattr(self, toolbar_name).__setstate__(storage)
            except (KeyError, TypeError) as e:
                logging.error(e)
                logging.warning(storage_err_msg(toolbar_name))



from __future__ import annotations

from PyQt6.QtGui import QColor, QPainter, QPaintEvent
from PyQt6.QtWidgets import QWidget

__all__ = [
    'ColorView'
]


class ColorView(QWidget):
    __slots__ = ('_color', )

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self._color = QColor(0, 0, 0, 255)

    def paintEvent(self, event: QPaintEvent) -> None:
        super().paintEvent(event)

        painter = QPainter(self)
        painter.fillRect(event.rect(), self.color)

    @property
    def color(self) -> QColor:
        return self._color

    @color.setter
    def color(self, value: QColor) -> None:
        if self._color == value:
            return
        self._color = value
        self.update()




from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Literal, cast

__all__ = [
    'install_vscode_launch'
]


def install_vscode_launch(mode: Literal['override', 'append', 'ignore'], path: str | Path | None = None) -> None:
    vscode_settings_path = Path(path or Path.cwd()) / '.vscode'
    vscode_settings_path.mkdir(0o777, True, True)

    common_args = {
        "type": "python",
        "request": "launch",
        "console": "internalConsole",
        "gevent": False,
        "justMyCode": True,
        "logToFile": False,
        "subProcess": False,
        "redirectOutput": True,
        "showReturnValue": False,
        "suppressMultipleSessionWarning": False
    }

    settings = {
        "version": "0.2.0",
        "configurations": [
            {
                "name": "VS Preview Current File",
                "module": "vspreview",
                "args": [
                    "${file}"
                ],
                **common_args  # type: ignore
            },
            {
                "name": "Run Current File",
                "program": "${file}",
                **common_args  # type: ignore
            }
        ]
    }

    launch = vscode_settings_path / 'launch.json'

    if launch.exists():
        if mode == 'ignore':
            return
    else:
        launch.touch()

    current_settings = settings.copy()

    def _write() -> None:
        with open(launch, 'w') as f:
            json.dump(current_settings, f, indent=4)
        return

    if mode != 'append':
        return _write()

    with open(launch, 'r+', encoding='utf-8') as f:
        try:
            current_settings = json.loads(f.read())
        except json.JSONDecodeError:
            current_settings = {}

    if 'configurations' not in current_settings or len(current_settings['configurations']) == 0:
        current_settings['configurations'] = settings['configurations']
        return _write()

    cast(list[Any], current_settings['configurations']).extend(settings['configurations'])

    current_settings['configurations'] = list({
        (
            '____' if row['name'] == 'VS Preview Current File' else
            ':'.join(str(row[column]) for column in row.keys())
        ): row
        for row in cast(list[dict[str, str]], current_settings['configurations'])
    }.values())

    return _write()




from __future__ import annotations

import inspect
from fractions import Fraction
from pathlib import Path
from typing import Any

from vapoursynth import AudioNode, RawNode, VideoNode, get_outputs

from .nodes import set_timecodes, update_node_info

__all__ = [
    'set_output'
]


def set_output(
    node: RawNode, name: str | bool | None = None,
    *, cache: bool = True, disable_comp: bool = False,
    timecodes: str | Path | dict[
        tuple[int | None, int | None], float | tuple[int, int] | Fraction
    ] | list[Fraction] | None = None, denominator: int = 1001,
    **kwargs: Any
) -> None:
    index = len(get_outputs())

    ref_id = str(id(node))

    if isinstance(node, VideoNode):
        title, node_type = 'Clip', VideoNode
    elif isinstance(node, AudioNode):
        title, node_type = 'Audio', AudioNode
    else:
        title, node_type = 'Node', RawNode

    if (not name and name is not False) or name is True:
        name = f"{title} {index}"

        current_frame = inspect.currentframe()

        assert current_frame
        assert current_frame.f_back

        for vname, val in reversed(current_frame.f_back.f_locals.items()):
            if (str(id(val)) == ref_id):
                name = vname
                break

    node.set_output(index)

    update_node_info(node_type, index, cache=cache, disable_comp=disable_comp, **kwargs)

    if name:
        update_node_info(node_type, index, name=name)

    if timecodes:
        set_timecodes(index, timecodes, node, denominator)



from __future__ import annotations

from typing import Any, Mapping

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QComboBox, QLabel
from vstools.functions.utils import DitherType

from ...core import AbstractToolbarSettings, ComboBox, Frame, HBoxLayout, SpinBox, try_load
from ...main import MainSettings
from ...models import GeneralModel

__all__ = [
    'PlaybackSettings'
]


class PlaybackSettings(AbstractToolbarSettings):
    __slots__ = ('buffer_size_spinbox', 'dither_type_combobox')

    CHECKERBOARD_ENABLED = True
    CHECKERBOARD_TILE_COLOR_1 = Qt.GlobalColor.white
    CHECKERBOARD_TILE_COLOR_2 = Qt.GlobalColor.lightGray
    CHECKERBOARD_TILE_SIZE = 8  # px
    FPS_AVERAGING_WINDOW_SIZE = Frame(100)
    FPS_REFRESH_INTERVAL = 150  # ms
    SEEK_STEP = 1

    def setup_ui(self) -> None:
        from ...core import main_window
        super().setup_ui()

        self.buffer_size_spinbox = SpinBox(self, 1, MainSettings.get_usable_cpus_count())
        self.dither_type_combobox = ComboBox[str](
            self, model=GeneralModel[str]([x.value for x in DitherType][1:]),
            currentIndex=3, sizeAdjustPolicy=QComboBox.SizeAdjustPolicy.AdjustToContents
        )

        self.dither_type_combobox.currentTextChanged.connect(lambda _: main_window().refresh_video_outputs())

        HBoxLayout(self.vlayout, [QLabel('Playback buffer size (frames)'), self.buffer_size_spinbox])
        HBoxLayout(self.vlayout, [QLabel('Dithering Type'), self.dither_type_combobox])

    def set_defaults(self) -> None:
        self.buffer_size_spinbox.setValue(MainSettings.get_usable_cpus_count())
        self.dither_type_combobox.setCurrentValue(DitherType.ERROR_DIFFUSION)

    @property
    def playback_buffer_size(self) -> int:
        return self.buffer_size_spinbox.value()

    @property
    def dither_type(self) -> str:
        return self.dither_type_combobox.currentValue()

    def __getstate__(self) -> Mapping[str, Any]:
        return super().__getstate__() | {
            'playback_buffer_size': self.playback_buffer_size, 'dither_type': self.dither_type
        }

    def _setstate_(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'playback_buffer_size', int, self.buffer_size_spinbox.setValue)
        try_load(state, 'dither_type', str, self.dither_type_combobox.setCurrentValue)



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Final, NamedTuple, cast

import vapoursynth as vs
from PyQt6 import QtCore
from PyQt6.QtCore import QKeyCombination, QObject, Qt, QThread, pyqtSignal
from PyQt6.QtWidgets import QComboBox, QLabel

from ...core import (
    AbstractToolbar, CheckBox, ComboBox, FrameEdit, HBoxLayout, LineEdit, ProgressBar, PushButton, VBoxLayout,
    VideoOutput, main_window
)
from ...models import GeneralModel
from .settings import CompSettings

if TYPE_CHECKING:
    from requests import Session

    from ...main import MainWindow


__all__ = [
    'CompToolbar'
]


_MAX_ATTEMPTS_PER_PICTURE_TYPE: Final[int] = 50


def _get_slowpic_headers(content_length: int, content_type: str, sess: Session) -> dict[str, str]:
    return {
        "Accept": "*/*",
        "Accept-Encoding": "gzip, deflate",
        "Accept-Language": "en-US,en;q=0.9",
        "Access-Control-Allow-Origin": "*",
        "Content-Length": str(content_length),
        "Content-Type": content_type,
        "Origin": "https://slow.pics/",
        "Referer": "https://slow.pics/comparison",
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36"
        ),
        "X-XSRF-TOKEN": sess.cookies.get("XSRF-TOKEN")
    }


def _do_single_slowpic_upload(sess: Session, collection: str, imageUuid: str, image: Path, browser_id: str) -> None:
    from uuid import uuid4

    from requests_toolbelt import MultipartEncoder  # type: ignore

    upload_info = MultipartEncoder({
        "collectionUuid": collection,
        "imageUuid": imageUuid,
        "file": (image.name, image.read_bytes(), 'image/png'),
        'browserId': browser_id,
    }, str(uuid4()))

    sess.post(
        'https://slow.pics/upload/image', data=upload_info.to_string(),
        headers=_get_slowpic_headers(upload_info.len, upload_info.content_type, sess)
    )


def clear_filename(filename: str) -> str:
    import re
    import unicodedata

    blacklist = ['\\', '/', ':', '*', '?', '\'', '<', '>', '|', '\0']
    reserved = [
        'CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5',
        'COM6', 'COM7', 'COM8', 'COM9', 'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5',
        'LPT6', 'LPT7', 'LPT8', 'LPT9',
    ]

    filename = ''.join(c for c in filename if c not in blacklist)

    # Remove all characters below code point 32
    filename = ''.join(c for c in filename if 31 < ord(c))
    filename = unicodedata.normalize('NFKD', filename).rstrip('. ').strip()

    if all([x == '.' for x in filename]):
        filename = '__' + filename

    if filename in reserved:
        filename = '__' + filename

    if len(filename) > 255:
        parts = re.split(r'/|\\', filename)[-1].split('.')

        if len(parts) > 1:
            ext = '.' + parts.pop()
            filename = filename[:-len(ext)]
        else:
            ext = ''
        if filename == '':
            filename = '__'

        if len(ext) > 254:
            ext = ext[254:]

        maxl = 255 - len(ext)
        filename = filename[:maxl]
        filename = filename + ext

        # Re-check last character (if there was no extension)
        filename = filename.rstrip('. ')

    return filename


class WorkerConfiguration(NamedTuple):
    uuid: str
    outputs: list[VideoOutput]
    collection_name: str
    public: bool
    nsfw: bool
    optimise: bool
    remove_after: int | None
    frames: list[int]
    compression: int
    path: Path
    main: MainWindow
    delete_cache: bool
    browser_id: str
    session_id: str
    tmdb: str
    tags: list[str]


class Worker(QObject):
    finished = pyqtSignal(str)
    progress_bar = pyqtSignal(str, int)
    progress_status = pyqtSignal(str, str, int, int)

    is_finished = False

    def _progress_update_func(self, value: int, endvalue: int, *, uuid: str) -> None:
        if value == 0:
            self.progress_bar.emit(uuid, 0)
        else:
            self.progress_bar.emit(uuid, int(100 * value / endvalue))

    def isFinished(self) -> bool:
        if self.is_finished:
            self.deleteLater()
        return self.is_finished

    def run(self, conf: WorkerConfiguration) -> None:
        import shutil
        from concurrent.futures import ThreadPoolExecutor
        from uuid import uuid4

        from requests import Session
        from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor  # type: ignore

        all_images = list[list[Path]]()
        conf.path.mkdir(parents=True, exist_ok=False)

        try:
            for i, output in enumerate(conf.outputs):
                if self.isFinished():
                    raise StopIteration
                self.progress_status.emit(conf.uuid, 'extract', i + 1, len(conf.outputs))

                folder_name = str(uuid4())
                path_name = conf.path / folder_name
                path_name.mkdir(parents=True)

                max_num = max(conf.frames)

                path_images = [
                    path_name / (f'{folder_name}_' + f'{f}'.zfill(len("%i" % max_num)) + '.png')
                    for f in conf.frames
                ]

                base_clip = output.prepared.clip

                if len(conf.frames) < 20:
                    decimated = vs.core.std.Splice([base_clip[i] for i in conf.frames])
                else:
                    decimated = output.prepared.clip.std.BlankClip(length=len(conf.frames))
                    decimated = decimated.std.FrameEval(lambda n: base_clip[conf.frames[n]])

                for i, f in enumerate(decimated.frames(close=True)):
                    if self.isFinished():
                        raise StopIteration

                    conf.main.current_output.frame_to_qimage(f).save(
                        str(path_images[i]), 'PNG', conf.compression
                    )

                    self._progress_update_func(i + 1, decimated.num_frames, uuid=conf.uuid)

                if self.isFinished():
                    raise StopIteration

                all_images.append(sorted(path_images))
        except StopIteration:
            return self.finished.emit(conf.uuid)
        except vs.Error as e:
            if 'raise StopIteration' in str(e):
                return self.finished.emit(conf.uuid)
            raise e

        total_images = 0
        fields = dict[str, Any]()
        for i, (output, images) in enumerate(zip(conf.outputs, all_images)):
            if self.isFinished():
                return self.finished.emit(conf.uuid)
            for j, (image, frame) in enumerate(zip(images, conf.frames)):
                if self.isFinished():
                    return self.finished.emit(conf.uuid)
                fields[f'comparisons[{j}].name'] = str(frame)
                fields[f'comparisons[{j}].imageNames[{i}]'] = output.name
                total_images += 1

        self.progress_status.emit(conf.uuid, 'upload', 0, 0)

        with Session() as sess:
            if conf.browser_id and conf.session_id:
                sess.cookies.set("SLPSESSION", conf.session_id, domain="slow.pics")
                browser_id = conf.browser_id
                check_session = True
            else:
                browser_id = str(uuid4())
                check_session = False

            base_page = sess.get('https://slow.pics/comparison')
            if self.isFinished():
                return self.finished.emit(conf.uuid)

            if check_session:
                if base_page.text.find('id="logoutBtn"') == -1:
                    self.progress_status.emit(conf.uuid, 'Session Expired', 0, 0)
                    return

            head_conf = {
                'collectionName': conf.collection_name,
                'hentai': str(conf.nsfw).lower(),
                'optimizeImages': str(conf.optimise).lower(),
                'browserId': browser_id,
                'public': str(conf.public).lower(),
            }
            if conf.remove_after is not None:
                head_conf |= {'removeAfter': str(conf.remove_after)}

            if conf.tmdb:
                head_conf |= {'tmdbId': conf.tmdb}

            if conf.public and conf.tags:
                head_conf |= {f'tags[{index}]': tag for index, tag in enumerate(conf.tags)}

            def _monitor_cb(monitor: MultipartEncoderMonitor) -> None:
                self._progress_update_func(monitor.bytes_read, monitor.len, uuid=conf.uuid)
            files = MultipartEncoder(head_conf | fields, str(uuid4()))
            monitor = MultipartEncoderMonitor(files, _monitor_cb)
            comp_response = sess.post(
                'https://slow.pics/upload/comparison', data=monitor.to_string(),
                headers=_get_slowpic_headers(monitor.len, monitor.content_type, sess)
            ).json()
            collection = comp_response["collectionUuid"]
            key = comp_response["key"]
            image_ids = comp_response["images"]
            images_done = 0
            self._progress_update_func(0, total_images, uuid=conf.uuid)
            with ThreadPoolExecutor() as executor:
                futures = []
                for i, (output, images) in enumerate(zip(conf.outputs, all_images)):
                    if self.isFinished():
                        return self.finished.emit(conf.uuid)
                    for j, (image, frame) in enumerate(zip(images, conf.frames)):
                        if self.isFinished():
                            return self.finished.emit(conf.uuid)
                        while len(futures) >= 5:
                            if self.isFinished():
                                return self.finished.emit(conf.uuid)
                            for future in futures.copy():
                                if self.isFinished():
                                    return self.finished.emit(conf.uuid)
                                if future.done():
                                    futures.remove(future)
                                    images_done += 1
                                    self._progress_update_func(
                                        images_done, total_images, uuid=conf.uuid
                                    )

                        futures.append(
                            executor.submit(
                                _do_single_slowpic_upload,
                                sess=sess, collection=collection, imageUuid=image_ids[j][i],
                                image=image, browser_id=browser_id
                            )
                        )
            self._progress_update_func(total_images, total_images, uuid=conf.uuid)
        if conf.delete_cache:
            shutil.rmtree(conf.path, True)

        url = f'https://slow.pics/c/{key}'

        self.progress_status.emit(conf.uuid, url, 0, 0)

        url_out = (
            conf.path.parent / 'Old Comps' / clear_filename(f'{conf.collection_name} - {key}')
        ).with_suffix('.url')
        url_out.parent.mkdir(parents=True, exist_ok=True)
        url_out.touch(exist_ok=True)
        url_out.write_text(f'[InternetShortcut]\nURL={url}')

        self.finished.emit(conf.uuid)


class CompToolbar(AbstractToolbar):
    _thread_running = False

    __slots__ = (
        'random_frames_control', 'manual_frames_lineedit', 'output_url_lineedit',
        'current_frame_checkbox', 'is_public_checkbox', 'is_nsfw_checkbox',
        'output_url_copy_button', 'start_upload_button', 'stop_upload_button',
        'upload_progressbar', 'upload_status_label', 'upload_status_elements'
    )

    settings: CompSettings

    upload_thread: QThread
    upload_worker: Worker

    curr_uuid = ''

    def __init__(self, main: MainWindow) -> None:
        super().__init__(main, CompSettings(self))
        self.setup_ui()

        self.set_qobject_names()

        self.add_shortcuts()

        self.tag_data_cache = None

        self.cache_state = None

    def setup_ui(self) -> None:
        super().setup_ui()

        self.collection_name_lineedit = LineEdit('Collection name', self)

        self.random_frames_control = FrameEdit(self)
        # self.bright_frames_control = FrameEdit(self)
        # self.dark_frames_control = FrameEdit(self)

        self.manual_frames_lineedit = LineEdit('Manual frames: frame,frame,frame', self, )

        self.current_frame_checkbox = CheckBox('Current frame', self, checked=True)

        def _force_clicked(self_s: str):
            def _on_clicked(is_checked: bool) -> None:
                if self_s == 'I':
                    el = self.pic_type_button_I
                    oth = (self.pic_type_button_P, self.pic_type_button_B)
                elif self_s == 'P':
                    el = self.pic_type_button_P
                    oth = (self.pic_type_button_I, self.pic_type_button_B)
                else:
                    el = self.pic_type_button_B
                    oth = (self.pic_type_button_I, self.pic_type_button_P)

                if not is_checked and not any(o.isChecked() for o in oth):
                    el.click()

            return _on_clicked

        def _select_filter_text(text: str):
            if text:
                if text in self.current_tags:
                    self.tag_add_custom_button.setText("Remove Tag")
                else:
                    self.tag_add_custom_button.setText("Add Tag")

                index = self.tag_list_combox.findText(text, QtCore.Qt.MatchFlag.MatchContains)
                if index < 0:
                    return
                self.tag_list_combox.setCurrentIndex(
                    index
                )

        def _handle_current_combox_tag():
            value = self.tag_list_combox.currentValue()
            if value in self.current_tags:
                self.current_tags.remove(value)
                self.tag_add_button.setText("Add Tag")
            else:
                self.current_tags.append(value)
                self.tag_add_button.setText("Remove Tag")

        def _handle_current_new_tag():
            value = self.tag_filter_lineedit.text()
            if value in self.current_tags:
                self.current_tags.remove(value)
                self.tag_add_custom_button.setText("Add Tag")
            else:
                self.current_tags.append(value)
                self.tag_add_custom_button.setText("Remove Tag")

        def _handle_tag_index(index):
            value = self.tag_list_combox.currentValue()
            if value in self.current_tags:
                self.tag_add_button.setText("Remove Tag")
            else:
                self.tag_add_button.setText("Add Tag")

        def _public_click(is_checked: bool):
            self.tag_add_button.setHidden(not is_checked)
            self.tag_add_custom_button.setHidden(not is_checked)
            self.tag_filter_lineedit.setHidden(not is_checked)
            self.tag_list_combox.setHidden(not is_checked)
            self.tag_separator.setHidden(not is_checked)

        self.pic_type_button_I = CheckBox('I', self, checked=True, clicked=_force_clicked('I'))
        self.pic_type_button_P = CheckBox('P', self, checked=True, clicked=_force_clicked('P'))
        self.pic_type_button_B = CheckBox('B', self, checked=True, clicked=_force_clicked('B'))

        self.tmdb_id_lineedit = LineEdit('TMDB ID', self)
        self.tmdb_id_lineedit.setMaximumWidth(75)

        self.tmdb_type_combox = ComboBox[str](
            self, model=GeneralModel[str](['TV', 'MOVIE'], to_title=False),
            currentIndex=0, sizeAdjustPolicy=QComboBox.SizeAdjustPolicy.AdjustToContents
        )

        self.tag_data = {"Error": "Error"}
        self.current_tags = []

        self.tag_list_combox = ComboBox[str](
            self, currentIndex=0, sizeAdjustPolicy=QComboBox.SizeAdjustPolicy.AdjustToContents
        )
        self.tag_list_combox.setMaximumWidth(250)

        self.update_tags()

        self.tag_filter_lineedit = LineEdit('Tag Selection', self, textChanged=_select_filter_text)

        self.tag_add_custom_button = PushButton('Add Tag', self, clicked=_handle_current_new_tag)

        self.tag_add_button = PushButton('Add Tag', self, clicked=_handle_current_combox_tag)

        self.tag_separator = self.get_separator()

        self.delete_after_lineedit = LineEdit('Days', self)
        self.delete_after_lineedit.setMaximumWidth(75)

        self.is_public_checkbox = CheckBox('Public', self, checked=False, clicked=_public_click)
        _public_click(self.is_public_checkbox.isChecked())

        self.is_nsfw_checkbox = CheckBox('NSFW', self, checked=False)

        self.output_url_lineedit = LineEdit('https://slow.pics/c/', self, enabled=False)

        self.output_url_copy_button = PushButton('⎘', self, clicked=self.on_copy_output_url_clicked)

        self.start_upload_button = PushButton('Start Upload', self, clicked=self.on_start_upload)

        self.stop_upload_button = PushButton('Stop Upload', self, visible=False, clicked=self.on_stop_upload)

        self.upload_progressbar = ProgressBar(self, value=0)
        self.upload_progressbar.setGeometry(200, 80, 250, 20)

        self.upload_status_label = QLabel(self, text='Select Frames')

        self.upload_status_elements = (self.upload_progressbar, self.upload_status_label)

        VBoxLayout(self.hlayout, [
            self.collection_name_lineedit,
            self.manual_frames_lineedit,
        ])

        self.collection_name_lineedit.setMinimumWidth(400)

        self.hlayout.addWidget(self.get_separator())

        VBoxLayout(self.hlayout, [
            HBoxLayout([
                self.current_frame_checkbox  # , self.get_separator(),
                # QLabel('Bright:'), self.bright_frames_control
            ]),
            HBoxLayout([
                QLabel('Random:'), self.random_frames_control,
                # QLabel('Dark:'), self.dark_frames_control,
            ])
        ])

        self.hlayout.addWidget(self.get_separator())

        VBoxLayout(self.hlayout, [
            QLabel('Picture types:'),
            HBoxLayout([
                self.pic_type_button_I,
                self.pic_type_button_P,
                self.pic_type_button_B
            ])
        ])

        self.hlayout.addWidget(self.get_separator())

        VBoxLayout(self.hlayout, [
            self.tmdb_id_lineedit,
            self.tmdb_type_combox,
        ])

        self.hlayout.addWidget(self.tag_separator)

        VBoxLayout(self.hlayout, [
            HBoxLayout([
                self.tag_filter_lineedit,
                self.tag_add_custom_button
            ]),
            HBoxLayout([
                self.tag_list_combox,
                self.tag_add_button,
            ])
        ])

        self.hlayout.addStretch(1)

        self.hlayout.addSpacing(20)
        VBoxLayout(self.hlayout, [
            self.is_public_checkbox,
            self.is_nsfw_checkbox,
        ])

        self.hlayout.addWidget(self.get_separator())

        self.output_url_lineedit.setMinimumWidth(250)

        VBoxLayout(self.hlayout, [
            HBoxLayout([
                self.output_url_lineedit, self.output_url_copy_button,
                QLabel('Delete After:'), self.delete_after_lineedit,
                self.start_upload_button, self.stop_upload_button
            ]),
            HBoxLayout([*self.upload_status_elements])
        ])

        self.tag_list_combox.currentIndexChanged.connect(_handle_tag_index)

    def add_shortcuts(self) -> None:
        self.main.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Space).toCombined(), self.add_current_frame_to_comp
        )

    def update_tags(self) -> None:
        self.tag_list_combox.setModel(GeneralModel[str](sorted(self.tag_data.keys()), to_title=False))

    def on_toggle(self, new_state: bool) -> None:
        if self.tag_data_cache is None or self.cache_state in ["ImportError", "Exception"]:
            try:
                from requests import Session

                with Session() as sess:
                    sess.get('https://slow.pics/comparison')

                    api_resp = sess.get("https://slow.pics/api/tags").json()

                    self.tag_data = {data["label"]: data["value"] for data in api_resp}

                    self.tag_data_cache = self.tag_data
                    self.cache_state = "Success"

            except ImportError:
                self.tag_data = {"Missing requests": "Missing requests"}
                self.cache_state = "ImportError"
            except Exception:
                self.tag_data = {"Network error": "Network error"}
                self.cache_state = "Exception"
        else:
            self.tag_data = self.tag_data_cache

        self.update_tags()

        super().on_toggle(new_state)

    def add_current_frame_to_comp(self) -> None:
        frame = str(self.main.current_output.last_showed_frame).strip()
        current_frames = self.manual_frames_lineedit.text()

        if not current_frames:
            self.manual_frames_lineedit.setText(frame)
        else:
            current_frames = current_frames.split(",")
            if frame not in current_frames:
                current_frames.append(frame)
            else:
                current_frames.remove(frame)
            self.manual_frames_lineedit.setText(",".join(current_frames))

    def on_copy_output_url_clicked(self, checked: bool | None = None) -> None:
        self.main.clipboard.setText(self.output_url_lineedit.text())
        self.main.show_message('Slow.pics URL copied to clipboard!')

    def on_start_upload(self) -> None:
        if self._thread_running:
            return
        if not self.upload_to_slowpics():
            return
        self.start_upload_button.setVisible(False)
        self.stop_upload_button.setVisible(True)

    def on_end_upload(self, uuid: str, forced: bool = False) -> None:
        if not forced and uuid != self.curr_uuid:
            return

        self.start_upload_button.setVisible(True)
        self.stop_upload_button.setVisible(False)
        self._thread_running = False

        if forced:
            self.upload_progressbar.setValue(int())
            self.upload_status_label.setText("Stopped!")
        else:
            self.upload_status_label.setText("Finished!")

    def on_stop_upload(self) -> None:
        self.upload_worker.is_finished = True

        self.on_end_upload(self.curr_uuid, forced=True)

    def update_status_label(self, uuid: str, kind: str, curr: int | None = None, total: int | None = None) -> None:
        if uuid != self.curr_uuid:
            return

        message = ''

        moreinfo = f" {curr or '?'}/{total or '?'} " if curr or total else ''

        if kind == 'extract':
            message = 'Extracting'
        elif kind == 'upload':
            message = 'Uploading'
        elif kind == 'search':
            message = 'Searching'
        else:
            self.output_url_lineedit.setText(kind)
            self.on_end_upload(self.curr_uuid, False)
            return

        self.upload_status_label.setText(f'{message}{moreinfo}...')

    def _rand_num_frames(self, checked: set[int], rand_func: Callable[[], int]) -> int:
        rnum = rand_func()
        while rnum in checked:
            rnum = rand_func()
        return rnum

    def _select_samples_ptypes(self, num_frames: int, k: int, picture_types: set[str]) -> list[int]:
        import logging
        import random

        samples = set[int]()
        _max_attempts = 0
        _rnum_checked = set[int]()

        assert self.main.outputs

        picture_types_b = {p.encode() for p in picture_types}

        while len(samples) < k:
            _attempts = 0
            while True:
                if self.upload_worker.is_finished:
                    raise RuntimeError

                self.update_status_label(self.curr_uuid, 'search', _attempts, _MAX_ATTEMPTS_PER_PICTURE_TYPE)
                if len(_rnum_checked) >= num_frames:
                    raise ValueError(f'There aren\'t enough of {picture_types} in these clips')
                rnum = self._rand_num_frames(_rnum_checked, partial(random.randrange, start=0, stop=num_frames))
                _rnum_checked.add(rnum)

                if all(
                    cast(bytes, f.props['_PictType']) in picture_types_b
                    for f in vs.core.std.Splice(
                        [out.prepared.clip[rnum] for out in self.main.outputs], True
                    ).frames(close=True)
                ):
                    break

                _attempts += 1
                _max_attempts += 1

                if _attempts > _MAX_ATTEMPTS_PER_PICTURE_TYPE:
                    logging.warning(
                        f'{_MAX_ATTEMPTS_PER_PICTURE_TYPE} attempts were made for sample {len(samples)} '
                        f'and no match found for {picture_types}; stopping iteration...')
                    break

            if _max_attempts > (curr_max_att := _MAX_ATTEMPTS_PER_PICTURE_TYPE * k):
                raise RecursionError(f'Comp: attempts max of {curr_max_att} has been reached!')

            if _attempts < _MAX_ATTEMPTS_PER_PICTURE_TYPE:
                samples.add(rnum)
                self.upload_progressbar.setValue(int())
                self.upload_progressbar.setValue(int(100 * len(samples) / k))

        return list(samples)

    def get_slowpics_conf(self) -> WorkerConfiguration:
        import logging
        import random
        import re
        import string
        from uuid import uuid4

        assert self.main.outputs

        num = int(self.random_frames_control.value())
        frames = list[int](
            map(int, filter(None, [x.strip() for x in self.manual_frames_lineedit.text().split(',')]))
        )
        picture_types = set[str]()

        if self.pic_type_button_I.isChecked():
            picture_types.add('I')

        if self.pic_type_button_B.isChecked():
            picture_types.add('B')

        if self.pic_type_button_P.isChecked():
            picture_types.add('P')

        lens = set(out.prepared.clip.num_frames for out in self.main.outputs)

        if len(lens) != 1:
            logging.warning('Outputted clips don\'t all have the same length!')

        lens_n = min(lens)

        path = Path(main_window().current_config_dir) / ''.join(
            random.choices(string.ascii_uppercase + string.digits, k=16)
        )

        if num:
            if picture_types == {'I', 'P', 'B'}:
                samples = random.sample(range(lens_n), num)
            else:
                logging.info('Making samples according to specified picture types...')
                samples = self._select_samples_ptypes(lens_n, num, picture_types)
        else:
            samples = []

        if len(frames):
            samples.extend(frames)

        if self.current_frame_checkbox.isChecked():
            samples.append(int(self.main.current_output.last_showed_frame))

        collection_name = self.collection_name_lineedit.text().strip()

        if not collection_name:
            collection_name = self.settings.DEFAULT_COLLECTION_NAME

        if not collection_name:
            raise ValueError('You have to put a collection name!')
        elif len(collection_name) <= 1:
            raise ValueError('Your collection name is too short!')

        collection_name = collection_name.format(
            script_name=self.main.script_path.stem
        )

        sample_frames = list(sorted(set(samples)))

        check_frame = sample_frames[0] if sample_frames else 0

        delete_after = self.delete_after_lineedit.text()
        if delete_after and re.match(r'^\d+$', delete_after) is None:
            raise ValueError('Delete after has to be a number!')
        elif not delete_after:
            delete_after = None

        tmdb_id = self.tmdb_id_lineedit.text()
        if tmdb_id:
            tmdb_type = self.tmdb_type_combox.currentData()
            if tmdb_type == "TV":
                suffix = "TV_"
            elif tmdb_type == "MOVIE":
                suffix = "MOVIE_"
            else:
                raise ValueError('Unknown TMDB type!')

            if not tmdb_id.startswith(suffix):
                tmdb_id = f"{suffix}{tmdb_id}"

        tags = [self.tag_data.get(tag, tag) for tag in self.current_tags]

        filtered_outputs = []
        for output in self.main.outputs:
            if output.info.get('disable_comp', False):
                continue

            filtered_outputs.append(output)

        return WorkerConfiguration(
            str(uuid4()), filtered_outputs, collection_name,
            self.is_public_checkbox.isChecked(), self.is_nsfw_checkbox.isChecked(),
            True, delete_after, sample_frames, -1, path, self.main, self.settings.delete_cache_enabled,
            self.settings.browser_id, self.settings.session_id, tmdb_id, tags
        )

    _old_threads_workers = list[Any]()

    def upload_to_slowpics(self) -> bool:
        try:
            self.main.current_scene.setPixmap(self.main.current_scene.pixmap().copy())

            if hasattr(self, 'upload_thread'):
                self._old_threads_workers.append(self.upload_thread)

            if hasattr(self, 'upload_worker'):
                self._old_threads_workers.append(self.upload_worker)

            self.upload_thread = QThread()
            self.upload_worker = Worker()

            self.upload_worker.moveToThread(self.upload_thread)

            try:
                config = self.get_slowpics_conf()
            except RuntimeError as e:
                print(e)
                return self.on_end_upload('', True)

            self.curr_uuid = config.uuid

            self.upload_thread.started.connect(partial(self.upload_worker.run, config))
            self.upload_thread.finished.connect(self.on_end_upload)

            self.upload_worker.progress_bar.connect(
                lambda uuid, val: self.upload_progressbar.setValue(val) if uuid == self.curr_uuid else None
            )
            self.upload_worker.progress_status.connect(self.update_status_label)

            self.upload_thread.start()

            self._thread_running = True

            return True
        except BaseException as e:
            self.main.show_message(str(e))

        return False



from __future__ import annotations

from abc import abstractmethod
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Mapping, Sequence, TypeAlias, cast, overload

from PyQt6.QtCore import QObject, Qt, QTimer, pyqtSignal
from PyQt6.QtGui import QKeySequence, QShortcut
from PyQt6.QtWidgets import (
    QApplication, QBoxLayout, QCheckBox, QDialog, QDoubleSpinBox, QFrame, QHBoxLayout, QLineEdit, QProgressBar,
    QPushButton, QSpacerItem, QSpinBox, QTableView, QVBoxLayout, QWidget
)

from .bases import QABC, QYAMLObjectSingleton

if TYPE_CHECKING:
    from vstools import T

    from ..main import MainWindow
    from .custom import Notches
    from .types import Frame, Stretch

    LayoutChildT: TypeAlias = QWidget | QBoxLayout | Stretch | QSpacerItem
else:
    LayoutChildT: TypeAlias = Any


__all__ = [
    'HBoxLayout', 'VBoxLayout',

    'SpinBox', 'PushButton', 'LineEdit', 'CheckBox', 'Timer', 'ProgressBar', 'DoubleSpinBox',

    'AbstractQItem', 'AbstractYAMLObject',

    'ExtendedWidgetBase', 'ExtendedWidget', 'ExtendedDialog', 'ExtendedTableView',

    'NotchProvider', 'AbstractToolbar', 'AbstractToolbarSettings',

    'main_window', 'storage_err_msg', 'try_load',
]


class ExtendedLayout(QBoxLayout):
    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, init_value: QWidget | QBoxLayout | None, **kwargs: Any) -> None:
        ...

    @overload
    def __init__(self, init_value: LayoutChildT | Sequence[LayoutChildT] | None, **kwargs: Any) -> None:
        ...

    @overload
    def __init__(
        self, parent: QWidget | QBoxLayout | None = None,
        children: LayoutChildT | Sequence[LayoutChildT] | None = None, **kwargs: Any
    ) -> None:
        ...

    def __init__(  # type: ignore
        self, arg0: QWidget | QBoxLayout | None = None,
        arg1: LayoutChildT | Sequence[LayoutChildT] | None = None,
        spacing: int | None = None, alignment: Qt.AlignmentFlag | None = None, **kwargs: Any
    ) -> ExtendedLayout:
        from .types import Stretch

        try:
            if isinstance(arg0, QBoxLayout):
                super().__init__(**kwargs)
                arg0.addLayout(self)
                arg0 = None
            elif isinstance(arg0, QWidget):
                super().__init__(arg0, **kwargs)  # type: ignore
                arg0 = None
            else:
                raise BaseException()
        except BaseException:
            super().__init__(**kwargs)

            if not any((arg0, arg1)):
                return  # type: ignore

        items = [u for s in (t if isinstance(t, Sequence) else [t] if t else [] for t in [arg0, arg1]) for u in s]

        for item in items:
            if isinstance(item, QBoxLayout):
                self.addLayout(item)
            elif isinstance(item, QSpacerItem):
                self.addSpacerItem(item)
            elif isinstance(item, Stretch):
                self.addStretch(item.amount)
            else:
                self.addWidget(item)

        for arg, action in ((spacing, 'setSpacing'), (alignment, 'setAlignment')):
            if arg is not None:
                getattr(self, action)(arg)

    def addWidgets(self, widgets: Sequence[QWidget]) -> None:
        for widget in widgets:
            self.addWidget(widget)

    def addLayouts(self, layouts: Sequence[QBoxLayout]) -> None:
        for layout in layouts:
            self.addLayout(layout)

    def clear(self) -> None:
        for i in reversed(range(self.count())):
            item = self.itemAt(i)
            widget = item.widget()
            self.removeItem(item)
            if widget:
                # removeItem only takes it out of the layout. The widget
                # still exists inside its parent widget.
                widget.deleteLater()
            else:
                # Clear and delete sub-layouts
                if isinstance(item, ExtendedLayout):
                    item.clear()
                item.deleteLater()  # type: ignore

    @staticmethod
    def stretch(amount: int | None) -> Stretch:  # type: ignore
        from .types import Stretch

        return Stretch(amount)  # type: ignore


class HBoxLayout(QHBoxLayout, ExtendedLayout):
    if TYPE_CHECKING:
        @overload
        def __init__(self) -> None:
            ...

        @overload
        def __init__(self, init_value: QWidget | QBoxLayout | None, **kwargs: Any) -> None:
            ...

        @overload
        def __init__(self, init_value: LayoutChildT | Sequence[LayoutChildT] | None, **kwargs: Any) -> None:
            ...

        @overload
        def __init__(
            self, parent: QWidget | QBoxLayout | None = None,
            children: LayoutChildT | Sequence[LayoutChildT] | None = None, **kwargs: Any
        ) -> None:
            ...

        def __init__(  # type: ignore
            self, arg0: QWidget | QBoxLayout | None = None,
            arg1: LayoutChildT | Sequence[LayoutChildT] | None = None,
            spacing: int | None = None, alignment: Qt.AlignmentFlag | None = None, **kwargs: Any
        ) -> ExtendedLayout:
            ...


class VBoxLayout(QVBoxLayout, ExtendedLayout):
    if TYPE_CHECKING:
        @overload
        def __init__(self) -> None:
            ...

        @overload
        def __init__(self, init_value: QWidget | QBoxLayout | None, **kwargs: Any) -> None:
            ...

        @overload
        def __init__(self, init_value: LayoutChildT | Sequence[LayoutChildT] | None, **kwargs: Any) -> None:
            ...

        @overload
        def __init__(
            self, parent: QWidget | QBoxLayout | None = None,
            children: LayoutChildT | Sequence[LayoutChildT] | None = None, **kwargs: Any
        ) -> None:
            ...

        def __init__(  # type: ignore
            self, arg0: QWidget | QBoxLayout | None = None,
            arg1: LayoutChildT | Sequence[LayoutChildT] | None = None,
            spacing: int | None = None, alignment: Qt.AlignmentFlag | None = None, **kwargs: Any
        ) -> ExtendedLayout:
            ...


class SpinBox(QSpinBox):
    def __init__(
        self, parent: QWidget | None = None, minimum: int | None = None,
        maximum: int | None = None, suffix: str | None = None, tooltip: str | None = None, **kwargs: Any
    ) -> None:
        super().__init__(parent, **kwargs)
        for arg, action in (
            (minimum, 'setMinimum'), (maximum, 'setMaximum'), (suffix, 'setSuffix'), (tooltip, 'setToolTip')
        ):
            if arg is not None:
                getattr(self, action)(arg)


if TYPE_CHECKING:
    ExtItemBase = QWidget
else:
    ExtItemBase = object


class ExtendedItemInit(ExtItemBase):
    def __init__(self, *args: QWidget | QBoxLayout | Stretch, tooltip: str | None = None, **kwargs: Any) -> None:
        try:
            super().__init__(*args, **kwargs)  # type: ignore
        except TypeError:
            super().__init__(*args)  # type: ignore

        if tooltip:
            super().setToolTip(tooltip)


class ExtendedItemWithName(ExtendedItemInit):
    if TYPE_CHECKING:
        def __init__(
            self, name: str, *args: QWidget | QBoxLayout | Stretch, tooltip: str | None = None, **kwargs: Any
        ) -> None:
            ...


class PushButton(ExtendedItemWithName, QPushButton):
    ...


class LineEdit(ExtendedItemInit, QLineEdit):
    def __init__(
        self, placeholder: str, *args: QWidget | QBoxLayout | Stretch, tooltip: str | None = None, **kwargs: Any
    ) -> None:
        return super().__init__(*args, tooltip=tooltip, **kwargs, placeholderText=placeholder)


class CheckBox(ExtendedItemWithName, QCheckBox):
    ...


class Timer(ExtendedItemInit, QTimer):
    ...


class ProgressBar(ExtendedItemInit, QProgressBar):
    ...


class DoubleSpinBox(ExtendedItemInit, QDoubleSpinBox):
    ...


class AbstractQItem:
    __slots__: tuple[str, ...]
    storable_attrs: ClassVar[tuple[str, ...]] = ()

    def add_shortcut(self, key: int, handler: Callable[[], None]) -> None:
        QShortcut(QKeySequence(key), self, activated=handler)  # type: ignore

    def set_qobject_names(self) -> None:
        if not hasattr(self, '__slots__'):
            return

        slots = list(self.__slots__)

        if isinstance(self, AbstractToolbar) and 'main' in slots:
            slots.remove('main')

        for attr_name in slots:
            attr = getattr(self, attr_name)
            if not isinstance(attr, QObject):
                continue
            attr.setObjectName(type(self).__name__ + '.' + attr_name)

    def __get_storable_attr__(self) -> tuple[str, ...]:
        return self.storable_attrs

    def __getstate__(self) -> Mapping[str, Any]:
        return {
            attr_name: getattr(self, attr_name)
            for attr_name in self.__get_storable_attr__()
        }


class AbstractYAMLObject(AbstractQItem):
    @abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError


class ExtendedWidgetBase(AbstractQItem):
    vlayout: VBoxLayout
    hlayout: HBoxLayout

    def setup_ui(self) -> None:
        self.vlayout = VBoxLayout(self)
        self.hlayout = HBoxLayout(self.vlayout)

    def add_shortcuts(self) -> None:
        pass

    def get_separator(self) -> QFrame:
        separator = QFrame(self)
        separator.setFrameShape(QFrame.Shape.VLine)
        separator.setFrameShadow(QFrame.Shadow.Sunken)
        return separator


class ExtendedWidget(ExtendedWidgetBase, QWidget):
    ...


class ExtendedDialog(AbstractQItem, QDialog):
    ...


class ExtendedTableView(AbstractQItem, QTableView):
    ...


class AbstractToolbarSettings(ExtendedWidget, QYAMLObjectSingleton):
    __slots__ = ()

    _add_to_tab = True

    def __init__(self, parent: type[AbstractToolbar] | AbstractToolbar) -> None:
        super().__init__()

        self.parent_toolbar_type = parent if isinstance(parent, type) else parent.__class__

        self.setup_ui()

        self.vlayout.addStretch(1)

        self.set_defaults()

        self.set_qobject_names()

    def set_defaults(self) -> None:
        pass

    def __getstate__(self) -> Mapping[str, Any]:
        return {}

    def _setstate_(self, state: Mapping[str, Any]) -> None:
        ...

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        if main_window().toolbars.should_set_state(self.__class__):
            self._setstate_(state)

            try:
                super().__setstate__(state)  # type: ignore
            except AttributeError:
                ...


class NotchProvider(QABC):
    notches_changed = pyqtSignal(ExtendedWidget)

    def init_notches(self, main: MainWindow = ...) -> None:
        self.notches_changed.connect(main.timeline.update_notches)

    def get_notches(self) -> Notches:
        from .custom import Notches
        return Notches()

    @property
    @abstractmethod
    def is_notches_visible(self) -> bool:
        ...


class AbstractToolbar(ExtendedWidget, NotchProvider):
    _no_visibility_choice = False
    storable_attrs = tuple[str, ...]()
    class_storable_attrs = tuple[str, ...](('settings', 'visibility'))
    num_keys = [
        Qt.Key.Key_1, Qt.Key.Key_2, Qt.Key.Key_3, Qt.Key.Key_4, Qt.Key.Key_5, Qt.Key.Key_6,
        Qt.Key.Key_7, Qt.Key.Key_8, Qt.Key.Key_9, Qt.Key.Key_0
    ]

    __slots__ = ('main', 'toggle_button', *class_storable_attrs)

    main: MainWindow
    name: str

    def __init__(self, main: MainWindow, settings: QWidget | None = None) -> None:
        super().__init__(main.central_widget)

        if settings is None:
            from vstools.exceptions import CustomValueError

            raise CustomValueError('Missing settings widget!')

        self.main = main
        self.settings = settings
        self.name = self.__class__.__name__[:-7]

        if settings._add_to_tab:
            self.main.app_settings.addTab(settings, self.name)
        self.setFocusPolicy(Qt.FocusPolicy.ClickFocus)

        self.toggle_button = PushButton(
            self.name, self, checkable=True, clicked=self.on_toggle
        )
        self.toggle_button.setVisible(not self._no_visibility_choice)

        self.setVisible(False)
        self.visibility = False

        self.init_notches(self.main)

    def setup_ui(self) -> None:
        self.hlayout = HBoxLayout(self)
        self.vlayout = VBoxLayout(self.hlayout)

        self.hlayout.setContentsMargins(0, 0, 0, 0)

    def on_toggle(self, new_state: bool) -> None:
        if new_state == self.visibility:
            return

        # invoking order matters
        self.setVisible(new_state)
        self.visibility = new_state
        self.toggle_button.setChecked(new_state)
        self.resize_main_window(new_state)

    def on_current_frame_changed(self, frame: Frame) -> None:
        pass

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        pass

    @property
    def is_notches_visible(self) -> bool:
        return self.visibility

    def resize_main_window(self, expanding: bool) -> None:
        if self.main.windowState() in {Qt.WindowState.WindowMaximized, Qt.WindowState.WindowFullScreen}:
            return

        if expanding:
            self.main.resize(self.main.width(), self.main.height() + self.height() + round(6 * self.main.display_scale))
        if not expanding:
            self.main.resize(self.main.width(), self.main.height() - self.height() - round(6 * self.main.display_scale))
            self.main.timeline.full_repaint()

    def __get_storable_attr__(self) -> tuple[str, ...]:
        attributes = list(self.class_storable_attrs + self.storable_attrs)

        if self._no_visibility_choice:
            attributes.remove('visibility')

        return tuple(attributes)

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        if not self._no_visibility_choice:
            try_load(state, 'visibility', bool, self.on_toggle)
        try_load(state, 'settings', AbstractToolbarSettings, self.__setattr__)


@lru_cache()
def main_window() -> MainWindow:
    import logging

    from ..main.window import MainWindow

    app = QApplication.instance()

    if app is not None:
        for widget in app.topLevelWidgets():  # type: ignore
            if isinstance(widget, MainWindow):
                return cast(MainWindow, widget)
        app.exit()

    logging.critical('main_window() failed')

    raise RuntimeError


class _OneArgumentFunction():
    def __call__(self, _arg0: T) -> Any:
        ...


class _SetterFunction():
    def __call__(self, _arg0: str, _arg1: T) -> Any:
        ...


def storage_err_msg(name: str, level: int = 0) -> str:
    import inspect

    pretty_name = name.replace('current_', ' ').replace('_enabled', ' ').replace('_', ' ').strip()
    caller_name = inspect.stack()[level + 1][0].f_locals['self'].__class__.__name__

    return f'Storage loading ({caller_name}): failed to parse {pretty_name}. Using default.'


def try_load(
    state: Mapping[str, Any], name: str, expected_type: type[T],
    receiver: T | _OneArgumentFunction | _SetterFunction,
    error_msg: str | None = None, nullable: bool = False
) -> None:
    import logging

    if error_msg is None:
        error_msg = storage_err_msg(name, 1)

    try:
        value = state[name]
        if not isinstance(value, expected_type) and not (nullable and value is None):
            raise TypeError
    except (KeyError, TypeError) as e:
        logging.error(e)
        logging.warning(error_msg)
        return
    finally:
        if nullable:
            value = None

    func_error = None

    if isinstance(receiver, expected_type):
        receiver = value
    elif callable(receiver):
        try:
            cast(_SetterFunction, receiver)(name, value)
        except Exception as e:
            if 'positional arguments but' not in str(e):
                func_error = e

            try:
                cast(_OneArgumentFunction, receiver)(value)
                func_error = None
            except Exception as ee:
                func_error = ee
    elif hasattr(receiver, name) and isinstance(getattr(receiver, name), expected_type):
        try:
            receiver.__setattr__(name, value)
        except AttributeError as e:
            logging.error(e)
            logging.warning(error_msg)

    if func_error:
        raise func_error from None



from __future__ import annotations

import inspect
import logging
import re
from functools import wraps
from time import perf_counter_ns
from typing import TYPE_CHECKING, Any, Callable, cast

from PyQt6 import sip
from PyQt6.QtCore import QEvent, QObject
from PyQt6.QtWidgets import QApplication, QGraphicsScene
from vstools import ColorRange, Matrix, Primaries, T, Transfer, vs

if TYPE_CHECKING:
    from ..main import MainWindow

__all__ = [
    'print_var',
    'print_func_name',
    'EventFilter',
    'measure_exec_time_ms',
    'print_perf_timepoints',
    'profile_cpu',
    'print_vs_output_colorspace_info',
    'DebugMeta',
    'GraphicsScene',
    'Application'
]


def print_var(var: Any) -> None:
    frame = inspect.currentframe()
    assert frame
    assert frame.f_back

    s = inspect.getframeinfo(frame.f_back).code_context
    assert s
    r = re.search(r"\((.*)\)", s[0])
    assert r
    logging.debug(f'{r.group(1)}: {var}')


def print_func_name() -> None:
    logging.debug(f'{inspect.stack()[1][3]}()')


class EventFilter(QObject):
    __slots__ = (
        'main',
    )

    def __init__(self, main: MainWindow) -> None:
        super().__init__()
        self.main = main

    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        if (event.type() == QEvent.Type.Show):
            logging.debug('--------------------------------')
            logging.debug(f'{obj.objectName()}')
            logging.debug('event:       Show')
            logging.debug(f'spontaneous: {event.spontaneous()}')
            logging.debug('')
            self.print_toolbars_state()
        elif (event.type() == QEvent.Type.Hide):
            logging.debug('--------------------------------')
            logging.debug(f'{obj.objectName()}')
            logging.debug('event:       Hide')
            logging.debug(f'spontaneous: {event.spontaneous()}')
            logging.debug('')
            self.print_toolbars_state()

        # return QObject.eventFilter(object, event)
        return False

    def print_toolbars_state(self) -> None:
        logging.debug(f'main toolbar:     {self.main.isVisible()}')
        if hasattr(self.main.toolbars, 'playback'):
            logging.debug(f'playback toolbar: {self.main.toolbars.playback.isVisible()}')
        if hasattr(self.main.toolbars, 'scening'):
            logging.debug(f'scening toolbar:  {self.main.toolbars.scening.isVisible()}')
        if hasattr(self.main.toolbars, 'misc'):
            logging.debug(f'misc toolbar:     {self.main.toolbars.misc.isVisible()}')

    def run_get_frame_test(self) -> None:
        N = 10

        start_frame_async = 1000
        total_async = 0
        for i in range(start_frame_async, start_frame_async + N):
            s1 = perf_counter_ns()
            f1 = self.main.current_output.prepared.clip.get_frame_async(i)
            f1.result()
            s2 = perf_counter_ns()
            logging.debug(f'async test time: {s2 - s1} ns')
            if i != start_frame_async:
                total_async += s2 - s1

        start_frame_sync = 2000
        total_sync = 0

        for i in range(start_frame_sync, start_frame_sync + N):
            s1 = perf_counter_ns()
            self.main.current_output.prepared.clip.get_frame(i)
            s2 = perf_counter_ns()
            if i != start_frame_sync:
                total_sync += s2 - s1

        logging.debug('')
        logging.debug(f'Async average: {total_async / N - 1} ns, {1_000_000_000 / (total_async / N - 1)} fps')
        logging.debug(f'Sync average:  {total_sync  / N - 1} ns, {1_000_000_000 / (total_sync  / N - 1)} fps')


def measure_exec_time_ms(
    func: Callable[..., T], return_exec_time: bool = False, print_exec_time: bool = True
) -> Callable[..., T | tuple[T, float]]:
    @wraps(func)
    def decorator(*args: Any, **kwargs: Any) -> T | tuple[T, float]:
        t1 = perf_counter_ns()
        ret = func(*args, **kwargs)
        t2 = perf_counter_ns()
        exec_time = (t2 - t1) / 1_000_000
        if print_exec_time:
            logging.debug(f'{exec_time:7.3f} ms: {func.__name__}()')
        if return_exec_time:
            return ret, exec_time
        return ret
    return decorator


def print_perf_timepoints(*args: int) -> None:
    if len(args) < 2:
        raise ValueError('At least 2 timepoints required')
    for i in range(1, len(args)):
        logging.debug(f'{i}: {args[i] - args[i-1]} ns')


def profile_cpu(func: Callable[..., T]) -> Callable[..., T]:
    @wraps(func)
    def decorator(*args: Any, **kwargs: Any) -> T:
        from cProfile import Profile
        from pstats import SortKey, Stats

        p = Profile(perf_counter_ns, 0.000_000_001, True, False)
        ret = p.runcall(func, *args, **kwargs)

        s = Stats(p)
        s.sort_stats(SortKey.TIME)
        s.print_stats(10)
        return ret
    return decorator


def print_vs_output_colorspace_info(vs_output: vs.VideoNode) -> None:
    props = vs_output.get_frame(0).props

    logging.debug('Matrix: {}, Transfer: {}, Primaries: {}, Range: {}'.format(
        Matrix.from_video(props) if '_Matrix' in props else None,
        Transfer.from_video(props) if '_Transfer' in props else None,
        Primaries.from_video(props) if '_Primaries' in props else None,
        ColorRange.from_video(props) if '_ColorRange' in props else None,
    ))


if TYPE_CHECKING:
    class DebugMetaBase(type, sip.wrappertype):
        ...
else:
    DebugMetaBase = sip.wrappertype


class DebugMeta(DebugMetaBase):
    def __new__(cls: type[type], name: str, bases: tuple[type, ...], dct: dict[str, Any]) -> type:
        from functools import partialmethod

        base = bases[0]

        # attr_list = [
        #     'activePanel', 'activeWindow', 'addEllipse', 'addItem', 'addLine', 'addPath', 'addPixmap',
        #     'addPolygon', 'addRect', 'addSimpleText', 'addText', 'addWidget', 'backgroundBrush',
        #     'bspTreeDepth', 'clearFocus', 'collidingItems', 'createItemGroup', 'destroyItemGroup',
        #     'focusItem', 'font', 'foregroundBrush', 'hasFocus', 'height', 'inputMethodQuery', 'invalidate',
        #     'isActive', 'itemAt', 'itemIndexMethod', 'items', 'itemsBoundingRect', 'minimumRenderSize',
        #     'palette', 'removeItem', 'render', 'sceneRect', 'selectedItems', 'selectionArea', 'sendEvent',
        #     'setActivePanel', 'setActiveWindow','setBackgroundBrush', 'setBspTreeDepth', 'setFocus',
        #     'setFocusItem', 'setFont', 'setForegroundBrush', 'setItemIndexMethod', 'setMinimumRenderSize',
        #     'setPalette', 'setSceneRect', 'setSelectionArea', 'setStickyFocus', 'setStyle', 'stickyFocus',
        #     'style', 'update', 'views', 'width', 'mouseGrabberItem'
        # ]

        for attr in dir(base):
            if not attr.endswith('__') and callable(getattr(base, attr)):
                dct[attr] = partialmethod(DebugMeta.dummy_method, attr)

        return super(DebugMeta, cls).__new__(cls, name, bases, dct)  # type: ignore

    def dummy_method(self, name: str, *args: Any, **kwargs: Any) -> Any:
        method = getattr(super(GraphicsScene, GraphicsScene), name)
        method = measure_exec_time_ms(method)
        return method(self, *args, **kwargs)


class GraphicsScene(QGraphicsScene, metaclass=DebugMeta):
    def event(self, event: QEvent) -> bool:
        t0 = perf_counter_ns()
        ret = super().event(event)
        t1 = perf_counter_ns()
        interval = t1 - t0
        if interval > 5_000_000:
            logging.debug(self.__class__.__name__ + '.event()')
            logging.debug(f'{interval / 1_000_000}: {event.type()}')

        return ret

    def __getattribute__(self, name: str) -> Any:
        attr = super().__getattribute__(name)
        if callable(attr):
            return measure_exec_time_ms(attr)


qevent_info = {
    0: ('None', 'invalid event'),
    1: ('Timer', 'timer event'),
    2: ('MouseButtonPress', 'mouse button pressed'),
    3: ('MouseButtonRelease', 'mouse button released'),
    4: ('MouseButtonDblClick', 'mouse button double click'),
    5: ('MouseMove', 'mouse move'),
    6: ('KeyPress', 'key pressed'),
    7: ('KeyRelease', 'key released'),
    8: ('FocusIn', 'keyboard focus received'),
    9: ('FocusOut', 'keyboard focus lost'),
    23: ('FocusAboutToChange', 'keyboard focus is about to be lost'),
    10: ('Enter', 'mouse enters widget'),
    11: ('Leave', 'mouse leaves widget'),
    12: ('Paint', 'paint widget'),
    13: ('Move', 'move widget'),
    14: ('Resize', 'resize widget'),
    15: ('Create', 'after widget creation'),
    16: ('Destroy', 'during widget destruction'),
    17: ('Show', 'widget is shown'),
    18: ('Hide', 'widget is hidden'),
    19: ('Close', 'request to close widget'),
    20: ('Quit', 'request to quit application'),
    21: ('ParentChange', 'widget has been reparented'),
    131: ('ParentAboutToChange', 'sent just before the parent change is done'),
    22: ('ThreadChange', 'object has changed threads'),
    24: ('WindowActivate', 'window was activated'),
    25: ('WindowDeactivate', 'window was deactivated'),
    26: ('ShowToParent', 'widget is shown to parent'),
    27: ('HideToParent', 'widget is hidden to parent'),
    31: ('Wheel', 'wheel event'),
    33: ('WindowTitleChange', 'window title changed'),
    34: ('WindowIconChange', 'icon changed'),
    35: ('ApplicationWindowIconChange', 'application icon changed'),
    36: ('ApplicationFontChange', 'application font changed'),
    37: ('ApplicationLayoutDirectionChange', 'application layout direction changed'),
    38: ('ApplicationPaletteChange', 'application palette changed'),
    39: ('PaletteChange', 'widget palette changed'),
    40: ('Clipboard', 'internal clipboard event'),
    42: ('Speech', 'reserved for speech input'),
    43: ('MetaCall', 'meta call event'),
    50: ('SockAct', 'socket activation'),
    132: ('WinEventAct', 'win event activation'),
    52: ('DeferredDelete', 'deferred delete event'),
    60: ('DragEnter', 'drag moves into widget'),
    61: ('DragMove', 'drag moves in widget'),
    62: ('DragLeave', 'drag leaves or is cancelled'),
    63: ('Drop', 'actual drop'),
    64: ('DragResponse', 'drag accepted/rejected'),
    68: ('ChildAdded', 'new child widget'),
    69: ('ChildPolished', 'polished child widget'),
    71: ('ChildRemoved', 'deleted child widget'),
    73: ('ShowWindowRequest', 'widget\'s window should be mapped'),
    74: ('PolishRequest', 'widget should be polished'),
    75: ('Polish', 'widget is polished'),
    76: ('LayoutRequest', 'widget should be relayouted'),
    77: ('UpdateRequest', 'widget should be repainted'),
    78: ('UpdateLater', 'request update() later'),

    79: ('EmbeddingControl', 'ActiveX embedding'),
    80: ('ActivateControl', 'ActiveX activation'),
    81: ('DeactivateControl', 'ActiveX deactivation'),
    82: ('ContextMenu', 'context popup menu'),
    83: ('InputMethod', 'input method'),
    87: ('TabletMove', 'Wacom tablet event'),
    88: ('LocaleChange', 'the system locale changed'),
    89: ('LanguageChange', 'the application language changed'),
    90: ('LayoutDirectionChange', 'the layout direction changed'),
    91: ('Style', 'internal style event'),
    92: ('TabletPress', 'tablet press'),
    93: ('TabletRelease', 'tablet release'),
    94: ('OkRequest', 'CE (Ok) button pressed'),
    95: ('HelpRequest', 'CE (?)  button pressed'),

    96: ('IconDrag', 'proxy icon dragged'),

    97: ('FontChange', 'font has changed'),
    98: ('EnabledChange', 'enabled state has changed'),
    99: ('ActivationChange', 'window activation has changed'),
    100: ('StyleChange', 'style has changed'),
    101: ('IconTextChange', 'icon text has changed.Deprecated.'),
    102: ('ModifiedChange', 'modified state has changed'),
    109: ('MouseTrackingChange', 'mouse tracking state has changed'),

    103: ('WindowBlocked', 'window is about to be blocked modally'),
    104: ('WindowUnblocked', 'windows modal blocking has ended'),
    105: ('WindowStateChange', ''),

    106: ('ReadOnlyChange', 'readonly state has changed'),

    110: ('ToolTip', ''),
    111: ('WhatsThis', ''),
    112: ('StatusTip', ''),

    113: ('ActionChanged', ''),
    114: ('ActionAdded', ''),
    115: ('ActionRemoved', ''),

    116: ('FileOpen', 'file open request'),

    117: ('Shortcut', 'shortcut triggered'),
    51: ('ShortcutOverride', 'shortcut override request'),

    118: ('WhatsThisClicked', ''),

    120: ('ToolBarChange', 'toolbar visibility toggled'),

    121: ('ApplicationActivate', 'deprecated. Use ApplicationStateChange instead.'),
    122: ('ApplicationDeactivate', 'deprecated. Use ApplicationStateChange instead.'),

    123: ('QueryWhatsThis', 'query what\'s this widget help'),
    124: ('EnterWhatsThisMode', ''),
    125: ('LeaveWhatsThisMode', ''),

    126: ('ZOrderChange', 'child widget has had its z-order changed'),

    127: ('HoverEnter', 'mouse cursor enters a hover widget'),
    128: ('HoverLeave', 'mouse cursor leaves a hover widget'),
    129: ('HoverMove', 'mouse cursor move inside a hover widget'),

    150: ('EnterEditFocus', 'enter edit mode in keypad navigation'),
    151: ('LeaveEditFocus', 'enter edit mode in keypad navigation'),
    152: ('AcceptDropsChange', ''),

    154: ('ZeroTimerEvent', 'Used for Windows Zero timer events'),

    155: ('GraphicsSceneMouseMove', 'GraphicsView'),
    156: ('GraphicsSceneMousePress', ''),
    157: ('GraphicsSceneMouseRelease', ''),
    158: ('GraphicsSceneMouseDoubleClick', ''),
    159: ('GraphicsSceneContextMenu', ''),
    160: ('GraphicsSceneHoverEnter', ''),
    161: ('GraphicsSceneHoverMove', ''),
    162: ('GraphicsSceneHoverLeave', ''),
    163: ('GraphicsSceneHelp', ''),
    164: ('GraphicsSceneDragEnter', ''),
    165: ('GraphicsSceneDragMove', ''),
    166: ('GraphicsSceneDragLeave', ''),
    167: ('GraphicsSceneDrop', ''),
    168: ('GraphicsSceneWheel', ''),

    169: ('KeyboardLayoutChange', 'keyboard layout changed'),

    170: ('DynamicPropertyChange', 'A dynamic property was changed through setProperty/property'),

    171: ('TabletEnterProximity', ''),
    172: ('TabletLeaveProximity', ''),

    173: ('NonClientAreaMouseMove', ''),
    174: ('NonClientAreaMouseButtonPress', ''),
    175: ('NonClientAreaMouseButtonRelease', ''),
    176: ('NonClientAreaMouseButtonDblClick', ''),

    177: ('MacSizeChange', 'when the Qt::WA_Mac{Normal,Small,Mini}Size changes'),

    178: ('ContentsRectChange', 'sent by QWidget::setContentsMargins (internal)'),

    179: ('MacGLWindowChange', 'Internal! the window of the GLWidget has changed'),

    180: ('FutureCallOut', ''),

    181: ('GraphicsSceneResize', ''),
    182: ('GraphicsSceneMove', ''),

    183: ('CursorChange', ''),
    184: ('ToolTipChange', ''),

    185: ('NetworkReplyUpdated', 'Internal for QNetworkReply'),

    186: ('GrabMouse', ''),
    187: ('UngrabMouse', ''),
    188: ('GrabKeyboard', ''),
    189: ('UngrabKeyboard', ''),
    191: ('MacGLClearDrawable', 'Internal Cocoa, the window has changed, so we must clear'),

    192: ('StateMachineSignal', ''),
    193: ('StateMachineWrapped', ''),

    194: ('TouchBegin', ''),
    195: ('TouchUpdate', ''),
    196: ('TouchEnd', ''),

    197: ('NativeGesture', 'QtGui native gesture'),
    199: ('RequestSoftwareInputPanel', ''),
    200: ('CloseSoftwareInputPanel', ''),

    203: ('WinIdChange', ''),
    198: ('Gesture', ''),
    202: ('GestureOverride', ''),
    204: ('ScrollPrepare', ''),
    205: ('Scroll', ''),

    206: ('Expose', ''),

    207: ('InputMethodQuery', ''),
    208: ('OrientationChange', 'Screen orientation has changed'),

    209: ('TouchCancel', ''),

    210: ('ThemeChange', ''),

    211: ('SockClose', 'socket closed'),

    212: ('PlatformPanel', ''),

    213: ('StyleAnimationUpdate', 'style animation target should be updated'),
    214: ('ApplicationStateChange', ''),

    215: ('WindowChangeInternal', 'internal for QQuickWidget'),
    216: ('ScreenChangeInternal', ''),

    217: ('PlatformSurface', 'Platform surface created or about to be destroyed'),

    218: ('Pointer', 'QQuickPointerEvent; ### Qt 6: QPointerEvent'),

    219: ('TabletTrackingChange', 'tablet tracking state has changed'),

    512: ('reserved', 'reserved for Qt Jambi\'s MetaCall event'),
    513: ('reserved', 'reserved for Qt Jambi\'s DeleteOnMainThread event'),

    1000: ('User', 'first user event id'),
    65535: ('MaxUser', 'last user event id'),
}


class Application(QApplication):
    enter_count = 0

    def notify(self, obj: QObject, event: QEvent) -> bool:
        import sys

        isex = False
        try:
            self.enter_count += 1
            ret, time = cast(tuple[bool, float], measure_exec_time_ms(
                QApplication.notify, True, False)(self, obj, event))
            self.enter_count -= 1

            if type(event).__name__ == 'QEvent' and event.type() in qevent_info:
                event_name = qevent_info[event.type()][0]
            else:
                event_name = type(event).__name__

            try:
                obj_name = obj.objectName()
            except RuntimeError:
                obj_name = ''

            if obj_name == '':
                try:
                    if obj.parent() is not None and obj.parent().objectName() != '':
                        obj_name = '(parent) ' + obj.parent().objectName()
                except RuntimeError:
                    pass

            recursive_indent = 2 * (self.enter_count - 1)

            logging.debug(
                f'{time:7.3f} ms, receiver: {type(obj).__name__:>25}, event: {event.type():3d} '
                f'{" " * recursive_indent + event_name:<30}, name: {obj_name}'
            )

            return ret
        except BaseException:
            isex = True
            logging.error('Application: unexpected error')
            logging.debug(*sys.exc_info())
            return False
        finally:
            if isex:
                self.quit()
